from .inspect import InspectContainer, Size
from .state import SaveContainer
from .env import EnvLoad, EnvHost, EnvSave
from .history import History
from .cp import Copy

# Based functions provided by paks
# These are currently all for docker and podman

# lookup of named commands and settings
docker_commands = {
    "#save": SaveContainer,
    "#inspect": InspectContainer,
    "#envload": EnvLoad,
    "#envhost": EnvHost,
    "#envsave": EnvSave,
    "#cp": Copy,
    "#size": Size,
}


class DockerCommands:

    # Required kwargs for any docker/podman command to run
    required = ["container_name", "name"]

    def __init__(self, container_tech):
        self.command = container_tech
        self.lookup = docker_commands

    def parse_name(self, cmd):
        parts = cmd.split(" ")
        return parts.pop(0).replace("\n", "").replace("\r", "").strip()

    def has_command(self, name):
        name, _ = self.parse_name(name)
        return name in self.lookup

    @property
    def history(self):
        return History(self.command)

    def get_executor(self, name, out=None):
        """
        Backend is required to update history
        """
        name = self.parse_name(name)
        if name in self.lookup:
            return self.lookup[name](self.command, required=self.required, out=out)



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

__version__ = "0.1.11"
AUTHOR = "Vanessa Sochat, Alec Scott"
EMAIL = "vsoch@noreply.github.users.com"
NAME = "paks"
PACKAGE_URL = "https://github.com/syspack/paks"
KEYWORDS = "containers, interactive commands."
DESCRIPTION = "Interactive container commands."
LICENSE = "LICENSE"

################################################################################
# Global requirements

INSTALL_REQUIRES = (
    ("pyaml", {"min_version": None}),
    ("jsonschema", {"min_version": None}),
    ("requests", {"min_version": None}),
)

TESTS_REQUIRES = (("pytest", {"min_version": "4.6.2"}),)

################################################################################
# Submodule Requirements (versions that include database)

INSTALL_REQUIRES_ALL = INSTALL_REQUIRES + TESTS_REQUIRES



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

from .command import Command
import json

# Every command must:
# 1. subclass Command
# 2. defined what container techs supported for (class attribute) defaults to all
# 3. define run function with kwargs


class Size(Command):

    supported_for = ["docker", "podman"]
    pre_message = "Sizing Container..."

    def run(self, **kwargs):
        """
        Get a container size.
        """
        # Always run this first to make sure container tech is valid
        self.check(**kwargs)

        # These are both required for docker/podman
        container_name = self.kwargs["container_name"]
        out, err = self.execute_host(
            [
                self.tech,
                "container",
                "ls",
                "-s",
                "--filter",
                "name=%s" % container_name,
                "--format",
                '"{{ json .}}"',
            ]
        )
        if not err:
            out = json.loads(out.strip().strip('"'))
            return self.return_success(out["Size"])

        return self.return_failure(err)


class InspectContainer(Command):

    supported_for = ["docker", "podman"]
    pre_message = "Inspecting Container..."

    def run(self, **kwargs):
        """
        Inspect a container fully, or specific sections
        """
        # Always run this first to make sure container tech is valid
        self.check(**kwargs)

        # These are both required for docker/podman
        container_name = self.kwargs["container_name"]

        # inspect defaults to labels and environment
        if self.args:
            for section in self.args:
                result = self.run_command(
                    [
                        self.tech,
                        "inspect",
                        "--format",
                        "{{json .%s }}" % section.capitalize(),
                        container_name,
                    ]
                )

        else:
            result = self.run_command([self.tech, "inspect", container_name])
            if result:
                return result
        return self.return_success()



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

from .command import Command
import os

# Every command must:
# 1. subclass Command
# 2. defined what container techs supported for (class attribute) defaults to all
# 3. define run function with kwargs


class EnvRemove(Command):

    supported_for = ["docker", "podman"]
    pre_message = "Saving environment variable..."
    parse_kwargs = False

    def run(self, **kwargs):
        """
        Save an environment variable to the host
        """
        self.check(**kwargs)

        if not self.args or len(self.args) < 2:
            return self.return_failure(
                "You must provide the environment name and at least one export."
            )
        envname = self.args.pop(0)
        self.env.load(envname)
        for envar in self.args:
            if not self.env.add(envname, envar, force=True):
                return self.return_failure(
                    "Could not add %s, did you include an =?" % envar
                )
            self.execute("export %s" % envar)
        return self.return_success(
            "Successfully added and exported environment variables."
        )


class EnvSave(Command):

    supported_for = ["docker", "podman"]
    pre_message = "Saving environment variable..."
    parse_kwargs = False

    def run(self, **kwargs):
        """
        Save an environment variable to the host
        """
        self.check(**kwargs)
        if not self.args or len(self.args) < 2:
            return self.return_failure(
                "You must provide the environment name and at least one export."
            )
        envname = self.args.pop(0)
        self.env.load(envname)
        for envar in self.args:
            if not self.env.add(envname, envar, force=True):
                return self.return_failure(
                    "Could not add %s, did you include an =?" % envar
                )
            self.execute("export %s" % envar)
        return self.return_success(
            "Successfully added and exported environment variables."
        )


class EnvHost(Command):

    supported_for = ["docker", "podman"]
    pre_message = "Getting host environment variable..."

    def run(self, **kwargs):
        """
        Retrieve an environment from the host.
        """
        # Always run this first to make sure container tech is valid
        self.check(**kwargs)

        # inspect defaults to labels and environment
        if not self.args:
            return self.return_failure(
                "You must provide the name of one or more environment variables."
            )

        # Load the environment variables
        found = False
        for name in self.args:
            value = os.environ.get(name)
            if not value:
                continue
            found = True
            self.execute("export %s=%s" % (name, value))

        if not found:
            return self.return_failure("No matching environment variables were found.")
        return self.return_success("Successfully loaded environment variables.")


class EnvLoad(Command):

    supported_for = ["docker", "podman"]
    pre_message = "Loading environment..."

    def run(self, **kwargs):
        """
        Load a named environment.
        """
        # Always run this first to make sure container tech is valid
        self.check(**kwargs)

        # inspect defaults to labels and environment
        if not self.args:
            return self.return_failure("You must provide the name of an environment.")

        # Load the environment!
        envname = self.args[0]
        self.env.load(envname)
        if not self.env.envars:
            return self.return_failure(
                "Environment %s does not have any variables." % envname
            )

        for key, value in self.env.envars.items():
            self.execute("export %s=%s" % (key, value))
        return self.return_success("Successfully loaded environment %s" % envname)



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

from paks.logger import logger
from pathlib import Path
import paks.defaults
import paks.utils
import time
import os


class Environment:
    def __init__(self, config_editor="vim", quiet=False):
        self.config_editor = config_editor
        self.envars = {}
        self.quiet = quiet

    def load(self, name):
        """
        Load an environtment by name
        """
        envpath = os.path.join(paks.defaults.paksenvs, name)
        self.ensure_exists(envpath)
        with open(envpath, "r") as fd:
            lines = fd.readlines()
        self.envars = {}
        for line in lines:
            key, value = self.parse_envar(line)
            self.envars[key] = value

    def save(self, envname):
        """
        Save environment by name
        """
        envpath = os.path.join(paks.defaults.paksenvs, envname)
        with open(envpath, "w") as fd:
            for key, value in self.envars.items():
                fd.write("%s=%s\n" % (key, value))

    def ensure_exists(self, envpath):
        """
        Ensure an envpath exists for an edit or load
        """
        if not os.path.exists(paks.defaults.paksenvs):
            paks.utils.mkdir_p(paks.defaults.paksenvs)
        envpath = os.path.join(paks.defaults.paksenvs, envpath)
        if not os.path.exists(envpath):

            # We have quiet mode for inside container commands
            if not self.quiet:
                logger.info("Creating %s" % envpath)
                time.sleep(0.5)
            Path(envpath).touch()

    def show(self, envpath):
        """
        show (print) a named environment to the terminal
        """
        envpath = os.path.join(paks.defaults.paksenvs, envpath)
        if not os.path.exists(envpath):
            logger.error("%s does not exist." % envpath)
        with open(envpath, "r") as fd:
            content = fd.read()
        return content

    def edit(self, envpath):
        """
        edit a named environment, or create if it doesn't exist.
        """
        self.ensure_exists(envpath)
        envpath = os.path.join(paks.defaults.paksenvs, envpath)

        # Make sure editor exists first!
        editor = paks.utils.which(self.config_editor)
        if not os.path.exists(editor):
            logger.exit(
                "Editor '%s' not found! Update with paks config set config_editor:<name>"
                % self.config_editor
            )
        os.system("%s %s" % (editor, envpath))

    def remove(self, envname, varname):
        """
        Remove a variable from a named environment.
        """
        envpath = os.path.join(paks.defaults.paksenvs, envname)
        if not os.path.exists(envpath):
            if not self.quiet:
                logger.error("%s does not exist." % envpath)
            return False
        self.load(envname)
        if varname in self.envars:
            del self.envars[varname]
        else:
            if not self.quiet:
                logger.error("%s is not found in %s" % (varname, envpath))
            return False
        self.save(envname)
        if not self.quiet:
            logger.info("%s was removed from environment %s" % (varname, envname))
        return True

    def parse_envar(self, param):
        """
        Parse a key value pair (ensure no export)
        """
        key, value = param.split("=", 1)

        # Ensure we remove export
        key = key.replace("export ", "", 1)
        return key.strip(), value.strip()

    def add(self, envname, keyval, force=False):
        """
        Add a variable to a named environment
        """
        if "=" not in keyval:
            logger.error("Expected key=value pair for variable")
            return False
        key, value = self.parse_envar(keyval)
        self.ensure_exists(envname)
        self.load(envname)
        if key in self.envars and not force:
            logger.error(
                "%s is already defined in %s, use --force to overwrite."
                % (key, envname)
            )
            return
        self.envars[key] = value
        self.save(envname)
        if not self.quiet:
            logger.info("%s was added to environment %s" % (key, envname))
        return True



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

import re

docker_regex = re.compile(
    "(?:(?P<registry>[^/@]+[.:][^/@]*)/)?"
    "(?P<namespace>(?:[^:@/]+/)+)?"
    "(?P<tool>[^:@/]+)"
    "(?::(?P<tag>[^:@]+))?"
    "(?:@(?P<digest>.+))?"
    "$"
)



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

from paks.client import PakClient
from paks.logger import logger


def main(args, parser, extra, subparser):
    cli = PakClient(args.settings_file)

    # Update settings
    for s in args.s or []:
        if "=" not in s:
            logger.warning("Malformed setting %s: skipping." % s)
            continue
        key, value = s.split("=", 1)
        cli.settings.add(key, value)

    # Run the container!
    cli.run(
        args.image,
        registry=args.registry,
        shell=args.shell,
        container_tech=args.container_tech,
    )



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

from .command import Command

# Every command must:
# 1. subclass Command
# 2. defined what container techs supported for (class attribute) defaults to all
# 3. define run function with kwargs


class Copy(Command):

    supported_for = ["docker", "podman"]
    pre_message = "Performing Copy..."

    def run(self, **kwargs):
        """
        Copy from host to container or vice versa
        """
        # Always run this first to make sure container tech is valid
        self.check(**kwargs)

        # These are both required for docker/podman
        container_name = self.kwargs["container_name"]

        # Args must have src and dst
        if len(self.args) != 2:
            return self.return_failure("You must provide a src and dest for copy.")

        src = self.args.pop(0)
        dest = self.args.pop(0)

        # src or dst must have host
        if "host:" not in src and "host:" not in dest:
            return self.return_failure(
                "One of copy arguments must be to the host:/path/to/file.txt"
            )

        args = {"src": src, "dest": dest}
        for argtype, path in args.items():
            if "host:" in path:
                path = path.replace("host:", "")
            else:
                path = "%s:%s" % (container_name, path)
            args[argtype] = path

        # docker cp!
        self.run_command([self.tech, "cp", args["src"], args["dest"]])
        return self.return_success()



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

from paks.logger import logger
import paks.commands
from .base import ContainerTechnology, ContainerName
import paks.utils

import subprocess


class DockerContainer(ContainerTechnology):
    """
    A Docker container controller.
    """

    command = "docker"

    def __init__(self, image, settings=None):
        if not paks.utils.which(self.command):
            logger.exit(
                "%s is required to use the '%s' base."
                % (self.command.capitalize(), self.command)
            )
        super(DockerContainer, self).__init__(settings)
        self.image = image
        self.uri = ContainerName(self.add_registry(image))
        self.commands = paks.commands.DockerCommands("docker")

    def run(self, shell):
        """
        Interactive shell into a container image.
        """
        cmd = [
            self.command,
            "run",
            "-it",
            "--rm",
            "--name",
            self.uri.extended_name,
            self.image,
            shell,
        ]
        name = self.interactive_command(cmd)

        # Remove the temporary container.
        if name:
            p = subprocess.Popen([self.command, "stop", name])
            p.wait()

    def add_registry(self, uri):
        """
        Given a "naked" name, add the registry if it's Docker Hub
        """
        # Is this a core library container, or Docker Hub without prefix?
        if uri.count("/") == 0:
            uri = "docker.io/library/%s" % uri
        elif uri.count("/") == 1:
            uri = "docker.io/%s" % uri
        return uri



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

from paks.utils.names import namer
from paks.logger import logger
import paks.utils
import paks.defaults
import paks.templates
import paks.commands
import paks.settings

import subprocess
import select
import string
import pty
import termios
import tty
import os
import sys
import re


class ContainerName:
    """
    Parse a container name into named parts
    """

    def __init__(self, raw):
        self.raw = raw
        self.registry = None
        self.repository = None
        self.tool = None
        self.version = None
        self.digest = None
        self.parse(raw)
        self._name = None

    @property
    def extended_name(self):
        if not self._name:
            self._name = namer.generate()
        return self.slug + "-" + self._name

    @property
    def name(self):
        if not self._name:
            self._name = namer.generate()
        return self._name

    @property
    def slug(self):
        slug = ""
        for attr in [self.registry, self.repository, self.tool, self.version]:
            if attr and not slug:
                slug = attr.replace(".", "")
            elif attr and slug:
                slug = slug + "-" + attr.replace(".", "")
        return slug

    def parse(self, raw):
        """
        Parse a name into known pieces
        """
        match = re.search(paks.templates.docker_regex, raw)
        if not match:
            logger.exit("%s does not match a known identifier pattern." % raw)
        for key, value in match.groupdict().items():
            value = value.strip("/") if value else None
            setattr(self, key, value)


class ContainerTechnology:
    """
    A base class for a container technology
    """

    def __init__(self, settings=None):
        if not settings:
            settings = paks.settings.Settings(paks.defaults.settings_file)
        self.settings = settings

    def get_history(self, line, openpty):
        """
        Given an input with some number of up/down and newline, derive command.
        """
        up = line.count("[A")
        down = line.count("[B")
        change = up - down

        # pushed down below history
        if change <= 0:
            return ""
        history = self.hist.run(
            container_name=self.uri.extended_name,
            out=openpty,
            history_file=self.settings.history_file,
            user=self.settings.user,
        )
        history = [x for x in history.split("\n") if x]

        if not history:
            return ""

        if change > len(history):
            return ""

        # here we are looking back up into history (negative index)
        newline = history[-1 * change]

        # Add back any characters typed
        newline += re.split("(\[A|\[B)", line, 1)[-1]
        return newline

    def encode(self, msg):
        return bytes((msg).encode("utf-8"))

    def interactive_command(self, cmd):
        """
        Ensure we always restore original TTY otherwise terminal gets messed up
        """
        # Controller to get history
        self.hist = self.commands.history

        # save original tty setting then set it to raw mode
        old_tty = termios.tcgetattr(sys.stdin)
        old_pty = termios.tcgetattr(sys.stdout)
        try:
            self._interactive_command(cmd)
        finally:
            termios.tcsetattr(sys.stdin, termios.TCSADRAIN, old_tty)
            termios.tcsetattr(sys.stdout, termios.TCSADRAIN, old_pty)

    def run_executor(self, string_input, openpty):
        """
        Given a string input, run executor
        """
        string_input = string_input.replace("[A", "").replace("[B", "")
        if not string_input.startswith("#"):
            return

        executor = self.commands.get_executor(string_input, out=openpty)
        if executor is not None:

            # Provide pre-command message to the terminal
            if executor.pre_message:
                print("\n\r" + executor.pre_message)

            # If we have an executor for the command, run it!
            # All commands require the original / current name
            result = executor.run(
                name=self.image,
                container_name=self.uri.extended_name,
                original=string_input,
            )
            if result.message:
                print("\r" + result.message)

    def clean(self, string_input):
        string_input = re.sub(
            r"[^a-zA-Z0-9%s\n\r\w ]" % string.punctuation, "", string_input
        )
        return string_input.replace("\x1b", "")

    def welcome(self, openpty):
        """
        Welcome the user and clear terminal
        """
        # Don't add commands executed to history
        os.write(openpty, self.encode(" export PROMPT_COMMAND='history -a'\r"))
        os.write(openpty, self.encode(" clear\r"))
        os.write(openpty, self.encode(" ### Welcome to PAKS! ###\r"))

    def _interactive_command(self, cmd):
        """
        Run an interactive command.
        """
        tty.setraw(sys.stdin.fileno())

        # open pseudo-terminal to interact with subprocess
        openpty, opentty = pty.openpty()

        # use os.setsid() make it run in a new process group, or bash job control will not be enabled
        p = subprocess.Popen(
            cmd,
            preexec_fn=os.setsid,
            stdin=opentty,
            stdout=opentty,
            stderr=opentty,
            universal_newlines=True,
        )

        # Welcome to Paks!
        self.welcome(openpty)
        string_input = ""

        while p.poll() is None:
            r, w, e = select.select([sys.stdin, openpty], [], [])
            if sys.stdin in r:
                terminal_input = os.read(sys.stdin.fileno(), 10240)
                new_char = terminal_input.decode("utf-8")

                # if we have a backspace (ord 127)
                if len(new_char) == 1 and ord(new_char) == 127:

                    # Backspace to empty line
                    if len(string_input) > 0:
                        string_input = string_input[:-1]
                    if not string_input:
                        os.write(openpty, terminal_input)
                        continue
                else:
                    string_input = string_input + new_char

                # Get rid of left/right
                string_input = string_input.replace("[D", "").replace("[C", "")
                has_newline = "\n" in string_input or "\r" in string_input

                # Replace weird characters and escape sequences
                string_input = self.clean(string_input)

                # Universal exit command
                if "exit" in string_input and has_newline:
                    print("\n\rContainer exited.\n\r")
                    return self.uri.extended_name

                # Pressing up or down, but not enter
                if ("[A" in string_input or "[B" in string_input) and not has_newline:
                    string_input = self.get_history(string_input, openpty)
                    os.write(openpty, terminal_input)
                    continue

                # Pressing up or down with enter
                if ("[A" in string_input or "[B" in string_input) and has_newline:
                    string_input = self.get_history(string_input, openpty)
                    os.write(openpty, terminal_input)

                if not string_input:
                    continue

                # If we have a newline (and possibly a command)
                if has_newline:
                    self.run_executor(string_input, openpty)

                    # Add derived line to the history
                    os.write(openpty, terminal_input)
                    string_input = ""
                else:
                    os.write(openpty, terminal_input)

            elif openpty in r:
                o = os.read(openpty, 10240)
                if o:
                    os.write(sys.stdout.fileno(), o)

    def __str__(self):
        return str(self.__class__.__name__)



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

from .command import Command

# Every command must:
# 1. subclass Command
# 2. defined what container techs supported for (class attribute) defaults to all
# 3. define run function with kwargs


class History(Command):

    supported_for = ["docker", "podman"]
    required = ["container_name"]

    def run(self, **kwargs):

        # Always run this first to make sure container tech is valid
        self.check(**kwargs)
        history_file = kwargs.get("history_file", "/root/.bash_history")
        self.out = self.kwargs.get("out", self.out)

        # These are both required for docker/podman
        container_name = self.kwargs["container_name"]
        out, err = self.execute_host(
            [
                self.tech,
                "exec",
                "-it",
                container_name,
                "/usr/bin/cat",
                history_file,
            ]
        )
        return out



from paks.logger import logger
from .podman import PodmanContainer
from .docker import DockerContainer


def get_container_backend(name, settings=None):
    techs = {
        "docker": DockerContainer,
        "podman": PodmanContainer,
    }
    if name.lower() not in techs:
        logger.exit("%s is not a known container backend." % name.lower())

    return techs[name.lower()]



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

import os
import locale
import subprocess
import shlex
import sys
import paks.env
import paks.utils


class Result:
    """
    A Result objet holds output and error, and a return value and message
    """

    def __init__(self, out=None, err=None, retval=0, msg=None):
        self.returncode = retval
        self.out = out or []
        self.err = err or []
        self.message = msg


class Command:
    """Class method to invoke shell commands and retrieve output and error.
    This class is inspired and derived from utils functions in
    https://github.com/vsoch/scif
    """

    # Assume we can support all three
    supported_tech = ["docker", "podman", "singularity"]

    # Message to print before run
    pre_message = None

    # Parse kwargs? (e.g., envars will have=)
    parse_kwargs = True

    def __init__(self, tech, required=None, out=None):
        """
        Backend is required to update history.
        """
        self.tech = tech
        self.required = required or []
        self.failed = False
        self.out = out or sys.stdout.fileno()

        # We don't need editor for interactive commands
        self.env = paks.env.Environment(quiet=True)

        # Don't add commands executed to history
        os.putenv("HISTCONTROL", "ignorespace")
        os.environ["HISTCONTROL"] = "ignorespace"

    def execute(self, cmd):
        """
        Execute a command to the container
        """
        # Extra space prevents saving to history
        os.write(self.out, self.encode(" \r %s" % cmd))

    def execute_get(self, runcmd, getcmd):
        """
        Execute and get runs a command inside the container (pipes to temporary
        file) and then loads from the outside.
        """
        # This is run inside the container
        self.run_hidden(runcmd)
        out, err = self.execute_host(getcmd)
        return out

    def execute_host(self, cmd):
        """
        Execute a command to the host, return out and error
        """
        # This is run outside the container
        res = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )
        out, err = res.communicate()
        return out, err

    def run_hidden(self, cmd):
        """
        Run a hidden command.
        """
        # TODO how to hide this?
        os.write(self.out, self.encode(" %s\r" % cmd))

    def check(self, **kwargs):
        """
        Check ensures that:

        1. The container tech of the command matches the class
        2. Required arguments are provided.
        """
        if self.tech not in self.supported_tech:
            return self.return_failure(
                "This command is not specified to run with %s." % self.tech
            )

        # Required args for all commands
        for entry in self.required:
            if entry not in kwargs:
                return self.return_failure("%s is required." % entry)

        args = kwargs.get("original")
        self.kwargs = kwargs

        # Get args (parsing from the original command line)
        parsed_args = []
        if args:
            parsed_args, kwargs = self.get_args(args)
            self.kwargs.update(kwargs)
        self.args = parsed_args

    def get_args(self, cmd):
        """
        Once we get here, we only care about additional command args.
        """
        parts = cmd.split(" ")

        # Pop off the command (we already use it)
        parts.pop(0).strip()
        kwargs = {}
        args = []
        for arg in parts:

            # This is an arg
            if "=" not in arg:
                arg = arg.strip()

                # Don't append empty args
                if not arg:
                    continue
                args.append(arg)
                continue

            # This is a kwarg
            if self.parse_kwargs:
                key, val = arg.split("=", 1)
                kwargs[key.strip()] = val.strip()

            # A command can choose to not split/parse
            else:
                args.append(arg.strip())
        return args, kwargs

    def return_failure(self, message, out=None, err=None):
        """
        Return a failed result (requires a message)
        """
        return Result(msg=message, retval=1, out=out, err=err)

    def return_success(self, message=None, out=None, err=None):
        """
        Return a successful result
        """
        return Result(msg=message, retval=0, out=None, err=None)

    def do_print(self, line, clear=True):
        if clear:
            print("\r")
        print(line, end="\r")

    def run_command(self, cmd, output="output"):
        """
        Wrapper to stream a command, which handles returning a result on error.
        """
        print("\r")
        lines = self.stream_command(cmd, output)
        while True:
            try:
                line = next(lines)
                self.do_print(line, False)

            # We use this to return the result
            except StopIteration as e:
                return e.value
                break

    def stream_command(self, cmd, output="output"):
        """
        Stream a command and use output or error.
        """
        process = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True
        )
        stream = process.stdout.readline
        if output == "error":
            stream = process.stderr.readline

        # Stream lines back to the caller
        for line in iter(stream, ""):
            yield line

        # If there is an error, raise.
        process.stdout.close()
        process.stderr.close()
        return_code = process.wait()

        # If failed, send failed result up to calling function
        if return_code:
            return self.return_failure("Failed: %s" % " ".join(cmd))

    def parse_command(self, cmd):
        """this is called when a new command is provided to ensure we have
        a list. We don't check that the executable is on the path,
        as the initialization might not occur in the runtime environment.
        """
        if not isinstance(cmd, list):
            cmd = shlex.split(cmd)
        return cmd

    def encode(self, line):
        return bytes(line.encode("utf-8"))

    def decode(self, line):
        """Given a line of output (error or regular) decode using the
        system default, if appropriate
        """
        loc = locale.getdefaultlocale()[1]

        try:
            line = line.decode(loc)
        except:
            pass
        return line



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

import importlib
import pkgutil

# A paks plugin (a pakage) can be discovered via install OR used on demand
# given that the user has dependencies needed installed.

installed_plugins = {
    name: importlib.import_module(name)
    for finder, name, ispkg in pkgutil.iter_modules()
    if name.startswith("pakage_")
}



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

from paks.backends import get_container_backend
from .settings import Settings
from .env import Environment


class PakClient:
    """
    Paks has a main controller for interacting with paks.
    """

    def __init__(self, *args, **kwargs):
        settings_file = kwargs.get("settings_file")
        validate = kwargs.get("validate", True)
        if not hasattr(self, "settings"):
            self.settings = Settings(settings_file, validate=validate)
        self.env = Environment(self.settings.config_editor)

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "[paks-client]"

    def run(self, image, registry=None, shell=None, container_tech=None):
        """
        Run a paks image
        """
        shell = shell or self.settings.container_shell
        backend = get_container_backend(container_tech or self.settings.container_tech)

        # Pass settings to the backend!
        backend(image, self.settings).run(shell)



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

from paks.utils.names import namer
from .command import Command
import shutil
import tempfile
import os

# Every command must:
# 1. subclass Command
# 2. defined what container techs supported for (class attribute) defaults to all
# 3. define run function with kwargs


class SaveContainer(Command):

    supported_for = ["docker", "podman"]
    pre_message = "Saving container..."

    def run(self, **kwargs):
        """
        Save a temporary container name back to the main container name
        Available after check - validated self.kwargs
        """
        # Always run this first to make sure container tech is valid
        self.check(**kwargs)

        # These are both required for docker/podman
        container_name = self.kwargs["container_name"]
        name = self.kwargs["name"]
        tmp_name = container_name + "-" + namer.generate()

        # Not required, so we have a default
        suffix = self.kwargs.get("suffix", "-saved")

        # Run the command (show in real time)
        result = self.run_command([self.tech, "commit", container_name, tmp_name])
        if result:
            return result

        # Keep track of where we are to change back to
        here = os.getcwd()

        # Create a temporary context
        tempdir = tempfile.mkdtemp()
        dockerfile = os.path.join(tempdir, "Dockerfile")
        with open(dockerfile, "w") as fd:
            fd.write("FROM %s\n" % tmp_name)
        os.chdir(tempdir)

        result = self.run_command(
            [self.tech, "build", "--squash", "-t", name + suffix, "."], "error"
        )
        if result:
            return result

        result = self.run_command([self.tech, "rmi", tmp_name])
        if result:
            return result

        # Remove dangling None images (not recommended lol)
        os.system(
            '%s rmi --force $(%s images --filter "dangling=true" -q --no-trunc) >/dev/null 2>&1'
            % (self.tech, self.tech)
        )
        os.chdir(here)
        shutil.rmtree(tempdir)
        return self.return_success("Successfully saved container! ⭐️")



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"


from .docker import DockerContainer

import os


class PodmanContainer(DockerContainer):
    """
    A Podman container controller, which is the same as Docker.
    """

    command = "podman"

    def shell(self, image):
        """
        Interactive shell into a container image.
        """
        os.system(
            "podman run -it --rm --entrypoint %s %s"
            % (self.settings.podman_shell, image)
        )



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

import os
import paks.utils as utils

install_dir = utils.get_installdir()
reps = {"$install_dir": install_dir, "$root_dir": os.path.dirname(install_dir)}

# The default settings file in the install root
default_settings_file = os.path.join(reps["$install_dir"], "settings.yml")

# Paks home stores user specific keys, etc
pakshome = os.path.expanduser("~/.paks")

# Paks environments
paksenvs = os.path.join(pakshome, "envs")

# The user settings file can be created to over-ride default
user_settings_file = os.path.join(pakshome, "settings.yml")

# Default directory for encrypted environments?
envs_dir = os.path.join(pakshome, "envs")

# Fallback content type
content_type = "application/vnd.paks.container"

# Allowed variables for the environment
allowed_envars = ["username"]

# Default registry prefix to install from
# TODO should we have some kind of pakages here meaning sets of commands?
trusted_packages_registry = "ghcr.io/pakages"

# The GitHub repository for paks
github_url = "https://github.com/syspack/paks"



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

from paks.client import PakClient
from paks.logger import logger
import sys


def main(args, parser, extra, subparser):
    """
    paks env edit <env-name>
    paks env add name=value <env-name>
    paks env remove name <env-name>
    paks env rm name <env-name>
    """
    # If nothing provided, show help
    if not args.params:
        print(subparser.format_help())
        sys.exit(0)

    # We are required a command and env name (minimally)
    if len(args.params) < 2:
        print(subparser.format_help())
        sys.exit(0)

    # The first "param" is either add, rm/remove, or edit
    command = args.params.pop(0)
    envname = args.params.pop(0)

    validate = True if not command == "edit" else False
    cli = PakClient(settings_file=args.settings_file, validate=validate)

    # For each new setting, update and save!
    if command == "edit":
        return cli.env.edit(envname)
    elif command == "show":
        content = cli.env.show(envname)
        print(content)
    elif command in ["rm", "remove"]:
        varname = envname
        if not args.params:
            logger.exit("Missing environment name at end.")
        envname = args.params.pop(0)
        cli.env.remove(envname, varname)
    elif command in ["add"]:
        keyval = envname
        if not args.params:
            logger.exit("Missing environment name at end.")
        envname = args.params.pop(0)
        cli.env.add(envname, keyval, force=args.force)
    else:
        logger.error("%s is not a recognized command." % command)



#!/usr/bin/env python

__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

import paks
from paks.logger import setup_logger
import argparse
import sys
import os


def get_parser():
    parser = argparse.ArgumentParser(
        description="paks interactive container commands",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    # Global Variables
    parser.add_argument(
        "--debug",
        dest="debug",
        help="use verbose logging to debug.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--quiet",
        dest="quiet",
        help="suppress additional output.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--version",
        dest="version",
        help="show software version.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--settings-file",
        dest="settings_file",
        help="custom path to settings file.",
    )
    subparsers = parser.add_subparsers(
        help="actions",
        title="actions",
        description="actions",
        dest="command",
    )

    # print version and exit
    subparsers.add_parser("version", description="show software version")

    run = subparsers.add_parser(
        "run",
        description="run an interactive container",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    run.add_argument(
        "--shell",
        "--sh",
        dest="shell",
        help="Use a one-off shell instead of the one defined in your config.",
    )
    run.add_argument(
        "-s", action="append", help="key=value argument to update settings."
    )

    # Paks environments
    env = subparsers.add_parser(
        "env",
        description="interact with paks environments",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    env.add_argument(
        "params",
        nargs="*",
        help="""interact with named environments written to ~/.paks/envs
paks env edit <env-name>
paks env add name=value <env-name>
paks env remove name <env-name>
paks env rm name <env-name>
""",
    )
    env.add_argument(
        "--force",
        dest="force",
        help="force replacement if environemnt variable already exists (for add)",
        default=False,
        action="store_true",
    )

    config = subparsers.add_parser(
        "config",
        description="update configuration settings. Use set or get to see or set information.",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    config.add_argument(
        "--central",
        "-c",
        dest="central",
        help="make edits to the central config file, if a user config is default.",
        default=False,
        action="store_true",
    )

    config.add_argument(
        "params",
        nargs="*",
        help="""Set or get a config value, edit the config, add or remove a list variable, or create a user-specific config.
paks config set key:value
paks config get key
paks edit
paks config init""",
        type=str,
    )

    for command in [run]:
        command.add_argument("image", help="image to run")

    for command in [run]:
        command.add_argument(
            "--registry",
            "-r",
            dest="registry",
            help="registry to use, if required for command.",
        )
        command.add_argument(
            "--container-tech",
            "-c",
            dest="container_tech",
            help="container technology to use, to over-ride user settings.",
        )

    return parser


def run_main():
    parser = get_parser()

    def help(return_code=0):
        """print help, including the software version and active client
        and exit with return code.
        """

        version = paks.__version__

        print("\nPaks v%s" % version)
        parser.print_help()
        sys.exit(return_code)

    # If the user didn't provide any arguments, show the full help
    if len(sys.argv) == 1:
        help()

    # If an error occurs while parsing the arguments, the interpreter will exit with value 2
    args, extra = parser.parse_known_args()

    if args.debug is True:
        os.environ["MESSAGELEVEL"] = "DEBUG"

    # Show the version and exit
    if args.command == "version" or args.version:
        print(paks.__version__)
        sys.exit(0)

    setup_logger(
        quiet=args.quiet,
        debug=args.debug,
    )

    # retrieve subparser (with help) from parser
    helper = None
    subparsers_actions = [
        action
        for action in parser._actions
        if isinstance(action, argparse._SubParsersAction)
    ]
    for subparsers_action in subparsers_actions:
        for choice, subparser in subparsers_action.choices.items():
            if choice == args.command:
                helper = subparser
                break

    if args.command == "run":
        from .run import main
    elif args.command == "config":
        from .config import main
    elif args.command == "env":
        from .env import main

    # Pass on to the correct parser
    return_code = 0
    try:
        main(args=args, parser=parser, extra=extra, subparser=helper)
        sys.exit(return_code)
    except UnboundLocalError:
        return_code = 1
    help(return_code)


if __name__ == "__main__":
    run_main()
